/* tapplay - TAP file playback for sd2iec hardware
   Copyright (C) 2009  Ingo Korb <ingo@akana.de>

   Inspiration and low-level SD/MMC access based on code from MMC2IEC
     by Lars Pontoppidan et al., see sdcard.c|h and config.h.

   FAT filesystem access based on code from ChaN and Jim Brain, see ff.c|h.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; version 2 of the License only.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


   tapsend.c: TAP playback code

*/

#include <avr/interrupt.h>
#include <avr/io.h>
#include <string.h>
#include <util/delay.h>
#include "config.h"
#include "avrcompat.h"
#include "diskio.h"
#include "ff.h"
#include "iec-ll.h"
#include "timer.h"
#include "uart.h"

// Dual-use buffer
#if _MAX_LFN_LENGTH > 255
uint8_t databuffer[_MAX_LFN_LENGTH+1];
#else
uint8_t databuffer[256];
#endif

static volatile uint8_t readidx;
static volatile uint8_t halfwave;
static uint8_t writeidx;
static uint8_t isractive;

static uint8_t longdelay;

ISR(TIMER0_COMPA_vect) {
  static uint16_t delay;
  uint8_t tmp;

  if (!longdelay) {
    tmp = databuffer[readidx];
    toggle_clock();
    toggle_srq();
    if (tmp) {
      OCR0A = (tmp) >> 1;
      if (halfwave) {
        halfwave = 0;
        readidx++;
      } else {
        halfwave = 1;
      }
    } else {
      readidx++;
      delay  = databuffer[++readidx];
      delay |= ((unsigned int)databuffer[++readidx] << 8);
      longdelay = 1;
      OCR0A = 0x1f;
    }
  } else {
    delay--;
    if (delay == 0) {
      longdelay = 0;
      set_clock(0);
      set_srq(1);
    }
  }
}

void send_file(FIL *fd) {
  OCR0A = 0xff;
  TCNT0 = 0;
  TCCR0A = _BV(WGM01);
  TCCR0B = _BV(CS01) | _BV(CS00);

  timer_disable();

  UINT bytesread;

  // Initialize with very short pulses in case the TAP file
  // is smaller than the buffer
  memset(databuffer,1,256);

  if (f_read(fd, databuffer, 256, &bytesread) != FR_OK) {
    dprintf("f_read failed\n");
    return;
  }

  uart_flush();

  // Start send-ISR
  set_data(0);
  readidx = 20; // Skip header
  halfwave = 0;
  set_clock(0);
  set_srq(1);
  _delay_ms(100);
  TIMSK0 |= _BV(OCIE0A);
  isractive = 1;

  writeidx = 0;
  while (bytesread > 0) {
    // Wait until ISR is in the new half of the buffer
    while ((readidx & 0x80) == (writeidx & 0x80)) ;

    f_read(fd, databuffer+writeidx, 128, &bytesread);
    writeidx += 128;

    if (uart_gotchar() && uart_getc() == 27)
      break;
  }

  // Wait once more (last read tried to read new half of buffer,
  // but failed -> need to wait until ISR is in new half, but
  // writeidx was incremented unconditionally -> wait until
  // ISR has left writeidx-half of buffer)
  while ((readidx & 0x80) == (writeidx & 0x80)) ;

  f_close(fd);

  timer_enable();

  TIMSK0 &= ~_BV(OCIE0A);

  set_data(1);
  set_clock(1);
  set_srq(1);
}
